"""Automatically generated by hassfest.

To update, run python3 -m script.hassfest
"""

# fmt: off

SSDP = {
    "deconz": [
        {
            "manufacturer": "Royal Philips Electronics"
        }
    ],
    "directv": [
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:1",
            "manufacturer": "DIRECTV"
        }
    ],
    "fritzbox": [
        {
            "st": "urn:schemas-upnp-org:device:fritzbox:1"
        }
    ],
    "harmony": [
        {
            "deviceType": "urn:myharmony-com:device:harmony:1",
            "manufacturer": "Logitech"
        }
    ],
    "heos": [
        {
            "st": "urn:schemas-denon-com:device:ACT-Denon:1"
        }
    ],
    "huawei_lte": [
        {
            "deviceType": "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
            "manufacturer": "Huawei"
        }
    ],
    "hue": [
        {
            "manufacturer": "Royal Philips Electronics",
            "modelName": "Philips hue bridge 2012"
        },
        {
            "manufacturer": "Royal Philips Electronics",
            "modelName": "Philips hue bridge 2015"
        },
        {
            "manufacturer": "Signify",
            "modelName": "Philips hue bridge 2015"
        }
    ],
    "konnected": [
        {
            "manufacturer": "konnected.io"
        }
    ],
    "roku": [
        {
            "deviceType": "urn:roku-com:device:player:1-0",
            "manufacturer": "Roku",
            "st": "roku:ecp"
        }
    ],
    "samsungtv": [
        {
            "st": "urn:samsung.com:device:RemoteControlReceiver:1"
        }
    ],
    "sonos": [
        {
            "st": "urn:schemas-upnp-org:device:ZonePlayer:1"
        }
    ],
    "synology_dsm": [
        {
            "deviceType": "urn:schemas-upnp-org:device:Basic:1",
            "manufacturer": "Synology"
        }
    ],
    "wemo": [
        {
            "manufacturer": "Belkin International Inc."
        }
    ]
}
