"""Constants for the Coronavirus integration."""
from coronavirus import DEFAULT_SOURCE

DOMAIN = "coronavirus"
OPTION_WORLDWIDE = "__worldwide"
ATTRIBUTION = f"Data provided by {DEFAULT_SOURCE.NAME}"
